/*
 * file_io.c -- file handling
 *
 * Copyright (C) Chris Ison 2001-2011
 * Copyright (C) Bret Curtis 2013-2016
 *
 * This file is part of WildMIDI.
 *
 * WildMIDI is free software: you can redistribute and/or modify the player
 * under the terms of the GNU General Public License and you can redistribute
 * and/or modify the library under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of
 * the licenses, or(at your option) any later version.
 *
 * WildMIDI is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License and
 * the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License and the
 * GNU Lesser General Public License along with WildMIDI.  If not,  see
 * <http://www.gnu.org/licenses/>.
 */

#ifndef __FILE_IO_H
#define __FILE_IO_H

#define WM_MAXFILESIZE 0x1fffffff
extern void *_WM_BufferFileImpl(const char *filename, uint32_t *size);
extern void  _WM_FreeBufferFileImpl(void*);
extern void * (*_WM_BufferFile)(const char *, uint32_t *);
extern void   (*_WM_FreeBufferFile)(void*);

#endif /* __FILE_IO_H */
